/* **********************************************************************************************************************
   Included Header Files
   **********************************************************************************************************************/

#include "VDDirect3D.h"



/* **********************************************************************************************************************
   Constructor
   **********************************************************************************************************************/

VDDirect3D::VDDirect3D() {
    direct3D = NULL;
    direct3DDevice = NULL;
    (void) SecureZeroMemory(&presentationParameters, sizeof(presentationParameters));
    sprites.clear();
    sprite = NULL;
    isMoving = FALSE;
    movingDirection = 0;
    movingDuration = 0;
};



/* **********************************************************************************************************************
   Destructor
   **********************************************************************************************************************/

VDDirect3D::~VDDirect3D() {
    if (font != NULL) {
        (void) font->Release();
        font = NULL;
    };
    if (sprite != NULL) {
        (void) sprite->Release();
        sprite = NULL;
    };
    if (direct3DDevice != NULL) {
        (void) direct3DDevice->Release();
        direct3DDevice = NULL;
    };
    if (direct3D != NULL) {
        (void) direct3D->Release();
        direct3D = NULL;
    };
};



/* **********************************************************************************************************************
   InitializeDirect3D
   ------------------
   Initializes and creates the Direct3D device.
   #param  window  Handle to the window.
   **********************************************************************************************************************/

void VDDirect3D::InitializeDirect3D(HWND window) {
    HRESULT returnValue = NULL;

    // create an IDirect3D9 object
    direct3D = Direct3DCreate9(D3D_SDK_VERSION);
    if (direct3D == NULL) {
        VDError("Failed to initialize Direct3D.", __FILE__, __LINE__);
    };

    // TODO: for full screen, use CheckDeviceType, GetDisplayMode, and EnumAdapterModes to
    //       find out the available options for full screen modes (16 bit, 32 bit, resolutions, etc.)

    // set up the presentation parameters
    presentationParameters.AutoDepthStencilFormat      = D3DFMT_UNKNOWN;          // format of automatic depth-stencil surface (ignored)
    presentationParameters.BackBufferCount             = 0;                       // number of back buffers
    presentationParameters.BackBufferFormat            = D3DFMT_UNKNOWN;          // back buffer format
    presentationParameters.BackBufferHeight            = 0;                       // height of swap chain's back buffer(s)
    presentationParameters.BackBufferWidth             = 0;                       // width of swap chain's back buffer(s)
    presentationParameters.EnableAutoDepthStencil      = FALSE;                   // Direct3D manages depth buffers
    presentationParameters.Flags;                                                 // various flag types
    //presentationParameters.FullScreen_RefreshRateInHz;
    //presentationParameters.hDeviceWindow;
    //presentationParameters.MultiSampleQuality;
    //presentationParameters.MultiSampleType;
    //presentationParameters.PresentationInterval;
    presentationParameters.SwapEffect                  = D3DSWAPEFFECT_DISCARD;
    presentationParameters.Windowed                    = TRUE;

    // create a device to represent the display adapter 
    returnValue = direct3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, window, D3DCREATE_HARDWARE_VERTEXPROCESSING, &presentationParameters, &direct3DDevice);
    //direct3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, window, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &presentationParameters, &direct3DDevice);
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to create the display adapter device.", returnValue, __FILE__, __LINE__);
    };

    // create a sprite object
    returnValue = D3DXCreateSprite(direct3DDevice, &sprite);
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to create the sprite object.", returnValue, __FILE__, __LINE__);
    };

    // create a font object
    returnValue = D3DXCreateFont(direct3DDevice, 16, 0, FW_NORMAL, 0, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, ANTIALIASED_QUALITY, DEFAULT_PITCH | FF_DONTCARE, "Arial", &font);
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to create the font object.", returnValue, __FILE__, __LINE__);
    };
};



/* **********************************************************************************************************************
   Render
   ------
   Renders the graphics to the screen.
   #param  area           A rectangle representing the area to draw.
   #param  windowWidth    Width of the window (used to offset the player in the main game, set to 0 for engine)
   #param  windowHeight   Height of the window (used to offset the player in the main game, set to 0 for engine)
   #param  userLocationX  The x location of the user.
   #param  userLocationY  The y location of the user.
   **********************************************************************************************************************/

void VDDirect3D::Render(const RECT *area, int windowWidth, int windowHeight, int userLocationX, int userLocationY) {
    HRESULT returnValue = NULL;

    // clear the screen
    returnValue = direct3DDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to clear render target surface.", returnValue, __FILE__, __LINE__);
    };

    // begin drawing
    (void) direct3DDevice->BeginScene();
    if (returnValue != D3D_OK) {
        VDDirectXError("IDirect3DDevice9::BeginScene is still running.  Make a call to EndScene first.", returnValue, __FILE__, __LINE__);
    };

        // are we in the process of moving?
        if (isMoving == TRUE && movingDuration < 32) {
            // traverse through the list of sprites twice
            for (list<VDSprite>::iterator aSprite = sprites.begin(); aSprite != sprites.end(); ++aSprite) {
                // begin drawing sprites
                returnValue = sprite->Begin(D3DXSPRITE_ALPHABLEND);
                if (returnValue != S_OK) {
                    VDDirectXError("Failed to prepare device for drawing sprites.", returnValue, __FILE__, __LINE__);
                };

                    // background sprites
                    if (aSprite->IsStatic() == FALSE && aSprite->IsNPC() == FALSE) {
                        // translate the sprites slowly
                        switch (movingDirection) {
                            case 1:
                                aSprite->ShiftLocation(direct3DDevice,
                                                       (float) (windowWidth / 2) - (userLocationX + 1) * 32.0f + (float) movingDuration,
                                                       (float) (windowHeight / 2) - userLocationY * 32.0f);
                                break;
                            case 2:
                                aSprite->ShiftLocation(direct3DDevice,
                                                       (float) (windowWidth / 2) - (userLocationX - 1) * 32.0f - (float) movingDuration,
                                                       (float) (windowHeight / 2) - userLocationY * 32.0f);
                                break;
                            case 3:
                                aSprite->ShiftLocation(direct3DDevice,
                                                       (float) (windowWidth / 2) - userLocationX * 32.0f,
                                                       (float) (windowHeight / 2) - (userLocationY + 1) * 32.0f + (float) movingDuration);
                                break;
                            case 4:
                                aSprite->ShiftLocation(direct3DDevice,
                                                       (float) (windowWidth / 2) - userLocationX * 32.0f,
                                                       (float) (windowHeight / 2) - (userLocationY - 1) * 32.0f - (float) movingDuration);
                                break;
                            default:
                                break;
                        };

                        // draw the sprites
                        aSprite->Draw(sprite);
                    };

                // end drawing sprites
                returnValue = sprite->End();
                if (returnValue != S_OK) {
                    VDDirectXError("Failed to end the drawing scene.", returnValue, __FILE__, __LINE__);
                };
            };
            for (list<VDSprite>::iterator aSprite = sprites.begin(); aSprite != sprites.end(); ++aSprite) {
                // begin drawing sprites
                returnValue = sprite->Begin(D3DXSPRITE_ALPHABLEND);
                if (returnValue != S_OK) {
                    VDDirectXError("Failed to prepare device for drawing sprites.", returnValue, __FILE__, __LINE__);
                };

                    // player/NPC sprites
                    if (aSprite->IsStatic() == TRUE || aSprite->IsNPC() == TRUE) {
                        // draw the sprites
                        aSprite->Draw(sprite);
                    };

                // end drawing sprites
                returnValue = sprite->End();
                if (returnValue != S_OK) {
                    VDDirectXError("Failed to end the drawing scene.", returnValue, __FILE__, __LINE__);
                };
            };

            // increment the moving process and see if we're finished
            movingDuration += 2;
            if (movingDuration == 32) {
                isMoving = FALSE;
                movingDirection = 0;
                movingDuration = 0;
            };

        // we're not moving
        } else {
            for (list<VDSprite>::iterator aSprite = sprites.begin(); aSprite != sprites.end(); ++aSprite) {
                // begin drawing sprites
                returnValue = sprite->Begin(D3DXSPRITE_ALPHABLEND);
                if (returnValue != S_OK) {
                    VDDirectXError("Failed to prepare device for drawing sprites.", returnValue, __FILE__, __LINE__);
                };

                    // background sprites
                    if (aSprite->IsStatic() == FALSE && aSprite->IsNPC() == FALSE) {
                        // translate the sprites according to the player's position
                        aSprite->ShiftLocation(direct3DDevice,
                                               (float) (windowWidth / 2) - userLocationX * 32.0f,
                                               (float) (windowHeight / 2) - userLocationY * 32.0f);

                        // draw the sprites
                        aSprite->Draw(sprite);
                    };

                // end drawing sprites
                returnValue = sprite->End();
                if (returnValue != S_OK) {
                    VDDirectXError("Invalid call to ID3DXSprite::End.", returnValue, __FILE__, __LINE__);
                };
            };
            for (list<VDSprite>::iterator aSprite = sprites.begin(); aSprite != sprites.end(); ++aSprite) {
                // begin drawing sprites
                returnValue = sprite->Begin(D3DXSPRITE_ALPHABLEND);
                if (returnValue != S_OK) {
                    VDDirectXError("Failed to prepare device for drawing sprites.", returnValue, __FILE__, __LINE__);
                };

                    // player/NPC sprites
                    if (aSprite->IsStatic() == TRUE || aSprite->IsNPC() == TRUE) {
                        // draw the static sprites
                        aSprite->Draw(sprite);
                    };

                // end drawing sprites
                returnValue = sprite->End();
                if (returnValue != S_OK) {
                    VDDirectXError("Invalid call to ID3DXSprite::End.", returnValue, __FILE__, __LINE__);
                };
            };
        };

        // draw text
        for (list<Text>::iterator someText = text.begin(); someText != text.end(); ++someText) {
            returnValue = font->DrawText(NULL, someText->message, -1, &someText->textBox, DT_LEFT | DT_NOCLIP, someText->color);
            if (returnValue == 0) {
                VDDirectXError("Failed to draw formatted text.", returnValue, __FILE__, __LINE__);
            };
        };

    // end drawing
    returnValue = direct3DDevice->EndScene();
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to end the drawing scene.", returnValue, __FILE__, __LINE__);
    };

    // present the content of the next buffer in the sequence of back buffers
    returnValue = direct3DDevice->Present(area, area, NULL, NULL);
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to present the display buffer.", returnValue, __FILE__, __LINE__);
    };
};



// TODO
void VDDirect3D::Render(const RECT *area, HWND window, VDSprite *aSprite) {
    HRESULT returnValue = NULL;

    // clear the screen
    returnValue = direct3DDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to clear render target surface.", returnValue, __FILE__, __LINE__);
    };

    // begin drawing
    (void) direct3DDevice->BeginScene();
    if (returnValue != D3D_OK) {
        VDDirectXError("IDirect3DDevice9::BeginScene is still running.  Make a call to EndScene first.", returnValue, __FILE__, __LINE__);
    };

        // begin drawing sprites
        returnValue = sprite->Begin(D3DXSPRITE_ALPHABLEND);
        if (returnValue != S_OK) {
            VDDirectXError("Failed to prepare device for drawing sprites.", returnValue, __FILE__, __LINE__);
        };

            // draw the sprite
            if (aSprite != NULL) {
                aSprite->Draw(sprite);
            };

        // end drawing sprites
        returnValue = sprite->End();
        if (returnValue != S_OK) {
            VDDirectXError("Invalid call to ID3DXSprite::End.", returnValue, __FILE__, __LINE__);
        };

    // end drawing
    returnValue = direct3DDevice->EndScene();
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to end the drawing scene.", returnValue, __FILE__, __LINE__);
    };

    // present the content of the next buffer in the sequence of back buffers
    returnValue = direct3DDevice->Present(area, area, window, NULL);
    if (returnValue != D3D_OK) {
        VDDirectXError("Failed to present the display buffer.", returnValue, __FILE__, __LINE__);
    };
};



/* **********************************************************************************************************************
   AddSprite
   ---------
   Creates a new 32x32 sprite and adds it to the list of sprites.
   #param  textureFile  Name of the texture (with directory structure included).
   #param  x            X coordinate of the sprite on the screen.
   #param  y            Y coordinate of the sprite on the screen.
   #param  width        Width of the sprite.
   #param  height       Height of the sprite.
   #param  isStatic     Does the sprite ever move?
   #param  isNPC        Is the sprite an NPC?
   #return  Pointer to the new sprite.
   **********************************************************************************************************************/

VDSprite* VDDirect3D::AddSprite(LPCTSTR textureFile, float x, float y, int width, int height, int isStatic, int isNPC) {
    // make sure the texture file was created
    if (strcmp(textureFile, "NULL") == 0) {
        return NULL;
    };

    VDSprite aSprite = VDSprite(direct3DDevice,
                                textureFile,
                                width,
                                height,
                                D3DXVECTOR3(0.0f, 0.0f, 0.0f),
                                D3DXVECTOR3(x, y, 0.0f),
                                D3DCOLOR_ARGB(255, 255, 255, 255),
                                isStatic,
                                isNPC);
    sprites.push_back(aSprite);

    return &sprites.back();
};



/* **********************************************************************************************************************
   RemoveTileSprite
   ----------------
   Looks for a tile in the given location and removes it if it exists.
   #param  x  X coordinate of the sprite.
   #param  y  Y coordinate of the sprite.
   **********************************************************************************************************************/

void VDDirect3D::RemoveTileSprite(float x, float y) {
    for (list<VDSprite>::iterator aSprite = sprites.begin(); aSprite != sprites.end();) {
        if (aSprite->GetPosition().x == x && aSprite->GetPosition().y == y) {
            if (aSprite->IsNPC() == FALSE) {
                aSprite = sprites.erase(aSprite);
            };
        } else {
            ++aSprite;
        };
    };
};



/* **********************************************************************************************************************
   RemoveSprites
   -------------
   Traverses through all of the sprites, removing any that are marked as "Dead".
   **********************************************************************************************************************/

void VDDirect3D::RemoveSprites() {
    for (list<VDSprite>::iterator aSprite = sprites.begin(); aSprite != sprites.end();) {
        if (aSprite->IsDead() == TRUE) {
            aSprite = sprites.erase(aSprite);
        } else {
            ++aSprite;
        };
    };
};



/* **********************************************************************************************************************
   MovingPlayer
   ------------
   Lets the renderer know that we're in the process of moving.
   #param  direction  The direction in which we wish to move.
   **********************************************************************************************************************/

void VDDirect3D::MovingPlayer(int direction) {
    isMoving = TRUE;
    movingDirection = direction;
};



/* **********************************************************************************************************************
   AddText
   -------
   Adds text to be rendered.
   #param  theMessage  The message of the text.
   #param  theLeft     Left X coordinate of the text box.
   #param  theTop      Top Y coordinate of the text box.
   #param  theRight    Right X coordinate of the text box.
   #param  theBottom   Bottom Y coordinate of the text box.
   #param  theColor    The color of the text.
   #return  A pointer to the first character of the message.  (TODO:  This may be unsafe)
   **********************************************************************************************************************/

char* VDDirect3D::AddText(char *theMessage, int theLeft, int theTop, int theRight, int theBottom, D3DCOLOR theColor) {
    Text newText;
    newText.textBox.left = theLeft;
    newText.textBox.top = theTop;
    newText.textBox.right = theRight;
    newText.textBox.bottom = theBottom;
    (void) strcpy_s(newText.message, theMessage);
    newText.color = theColor;
    text.push_back(newText);

    return text.back().message;
};



/* **********************************************************************************************************************
   RemoveText
   ----------
   Removes text from the screen that is the exact given string.
   #param  theMessage  The text to remove.
   **********************************************************************************************************************/

void VDDirect3D::RemoveText(char *theMessage) {
    for (list<Text>::iterator someText = text.begin(); someText != text.end();) {
        if (strcmp(theMessage, someText->message) == 0) {
            someText = text.erase(someText);
        } else {
            ++someText;
        };
    };
};



/* **********************************************************************************************************************
   RemoveTexts
   -----------
   Removes all text.
   **********************************************************************************************************************/

void VDDirect3D::RemoveTexts() {
    for (list<Text>::iterator someText = text.begin(); someText != text.end();) {
        someText = text.erase(someText);
    };
};



/* **********************************************************************************************************************
   Accessors and Modifiers
   **********************************************************************************************************************/

LPDIRECT3DDEVICE9 VDDirect3D::GetDirect3DRenderer() {
    return direct3DDevice;
};

LPD3DXSPRITE VDDirect3D::GetSpriteRenderer() {
    return sprite;
};

int VDDirect3D::IsMoving() {
    return isMoving;
};