/* **********************************************************************************************************************
   Included Header Files
   **********************************************************************************************************************/

#include "VDEngine.h"



/* **********************************************************************************************************************
   Constructor
   **********************************************************************************************************************/

VDEngine::VDEngine() {
    mapMakerRunning = TRUE;
};



/* **********************************************************************************************************************
   Destructor
   **********************************************************************************************************************/

VDEngine::~VDEngine() {
};



/* **********************************************************************************************************************
   Global Variables
   **********************************************************************************************************************/

// these are used to keep track of whether the windows are open or not
HWND vdEngineWindow = NULL;
HWND mapEditorWindow = NULL;
HWND npcEditorWindow = NULL;



/* **********************************************************************************************************************
   WindowProcedure
   ---------------
   Processes messages for the main window.
   **********************************************************************************************************************/

LRESULT CALLBACK WindowProcedure(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam) {
    switch (Msg) {
        case WM_CLOSE:
            // user clicked "X"
            DestroyWindow(hWnd);
            if (hWnd == mapEditorWindow) {
                mapEditorWindow = NULL;
            } else if (hWnd == npcEditorWindow) {
                npcEditorWindow = NULL;
            };
            break;
        case WM_DESTROY:
            // sent by WM_CLOSE, only exit if it's the main window clicking "X"
            if (hWnd == vdEngineWindow) {
                PostQuitMessage(0);
            } else {
                hWnd = NULL;
            };
            break;
        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                // user selected Exit from menu
                case VD_MENU_FILE_EXIT:
                    PostQuitMessage(0);
                    break;
                case VD_MENU_HELP_ABOUT:
                    (void) MessageBox(NULL, "Varkalas D`Lonovan's 2D Game Engine", "VDEngine", MB_OK);
                    break;
                default:
                    break;
            };
            break;
        default:
            break;
    };
    return DefWindowProc(hWnd, Msg, wParam, lParam);
};



/* **********************************************************************************************************************
   Run
   ---
   Main function of the map maker.
   #param  instance  The instance of the program.
   **********************************************************************************************************************/

void VDEngine::Run(HINSTANCE instance) {
    VDMapEditor mapEditor;
    VDNPCEditor npcEditor;
    VDDirect3D graphics;
    VDInput input;
    VDWorld world;
    VDEntity user;
    VDEntity worldOffset;
    MSG message;
    int width, height;
    int choice;
    POINT mousePosition;
    RECT windowDimensions;
    RECT renderDimensions;
    char mapEditorSpriteFileName[FILE_NAME_MAX_SIZE];
    char npcEditorPreviousNPCSelection[FILE_NAME_MAX_SIZE];
    char npcEditorPreviousDialogueSelection[FILE_NAME_MAX_SIZE];

    // initialize variables
    width = 640;
    height = 480;
    user.SetCurrentLocation(0, 0);
    worldOffset.SetCurrentLocation(0, 0);
    choice = 0;
    (void) strcpy_s(mapEditorSpriteFileName, "");
    (void) strcpy_s(npcEditorPreviousNPCSelection, "");
    (void) strcpy_s(npcEditorPreviousDialogueSelection, "");

    // register all of the windows
    RegisterWindows(instance);

    // create the application window
    vdEngineWindow = CreateVDEngineWindow(instance, VD_ENGINE_CLASS_NAME, width, height);
    if (vdEngineWindow == NULL) {
        VDError("Failed to create the window.", __FILE__, __LINE__);
    };

    // initialize the graphics API
    graphics.InitializeDirect3D(vdEngineWindow);

    // initialize input
    input.InitializeInput(instance);
    input.InitializeKeyboard(vdEngineWindow);
    input.InitializeMouse(vdEngineWindow);

    // main loop
    while (mapMakerRunning) {
        // dispatch incoming sent messages, check the thread message queue, retrieve message
        while (PeekMessage(&message, NULL, 0, 0, PM_REMOVE)) {
            if (message.message == WM_QUIT) {
                return;
            } else if (message.message == WM_COMMAND) {
                // now we shouldn't be checking this here... but it's the best way to open things from the menu
                switch (LOWORD(message.wParam)) {
                    case VD_MENU_FILE_MAP_EDITOR:
                        // load the Map Editor interface
                        if (mapEditorWindow == NULL) {
                            mapEditorWindow = mapEditor.CreateMapEditor(instance, VD_MAP_EDITOR_CLASS_NAME);
                        };
                        break;
                    case VD_MENU_FILE_NPC_EDITOR:
                        // load the NPC Editor interface
                        if (npcEditorWindow == NULL) {
                            npcEditorWindow = npcEditor.CreateNPCEditor(instance, VD_NPC_EDITOR_CLASS_NAME, world.GetNPCs());
                        };
                        break;
                    default:
                        break;
                };
            };
            TranslateMessage(&message);
            DispatchMessage(&message);
        };

        // gather input
        input.RetrieveLostInput();
        input.UpdateInput();
        GetCursorPos(&mousePosition);

        // take a look at the current window location
        (void) GetWindowRect(vdEngineWindow, &windowDimensions);

        // acquire the real mouse position
        int actualMousePositionX = mousePosition.x - GetSystemMetrics(SM_CXBORDER) - windowDimensions.left;
        int actualMousePositionY = mousePosition.y - (GetSystemMetrics(SM_CYBORDER) + GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU)) - windowDimensions.top;



        // Map Editor interface
        ///////////////////////

        if (mapEditor.GetWindow() != NULL) {
            // Save Map button
            if (SendMessage(mapEditor.GetSaveMapButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];

                // make sure there's a name in the edit box field
                (void) GetWindowText(mapEditor.GetMapNameEditBox(), text, sizeof(text));
                if (strcmp(text, "") != 0) {
                    // append .vd to the name and save the world
                    (void) strcat_s(text, ".vd");
                    world.CreateWorld(text);
                } else {
                    MessageBox(vdEngineWindow, "No map name was provided.", "VDEngine", MB_OK);
                };

                // make sure we're not still focused on the button
                SetFocus(mapEditor.GetWindow());
            };

            // Load Map button
            if (SendMessage(mapEditor.GetLoadMapButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];

                // open a world file
                VDOpenFile(mapEditor.GetWindow(), "VDEngine Maps\0*.vd\0", text);

                // make sure the user selected a file
                if (strcmp(text, "") != 0) {
                    // load the world
                    // TODO:  Need to load in NPC's separately.  That'll involve changing CreateWorld as well as LoadWorld
                    world.LoadWorld(text);

                    // add the tile sprites
                    for (int x = 0; x < world.GetWorldSizeX(); x++) {
                        for (int y = 0; y < world.GetWorldSizeY(); y++) {
                            graphics.AddSprite(world.GetTileImageFileLocation(x, y),
                                               (float) x * 32.0f,
                                               (float) y * 32.0f,
                                               32,
                                               32,
                                               FALSE,
                                               FALSE);
                        };
                    };

                    // add the NPC sprites
                };

                // make sure we're not still focused on the button
                SetFocus(mapEditor.GetWindow());
            };

            // Load Sprite button
            if (SendMessage(mapEditor.GetLoadSpriteButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                RECT drawingArea;

                // open a sprite image
                VDOpenFile(mapEditor.GetWindow(), "Sprites\0*.PNG\0All Files\0*.*\0", mapEditorSpriteFileName);

                // make sure the user selected a file
                if (strcmp(mapEditorSpriteFileName, "") != 0) {
                    // get the sprite's dimensions to pass into VDSprite()
                    (void) mapEditor.GetTileSpriteDimensions(&drawingArea);

                    // add the sprite to the editor so it may be drawn
                    mapEditor.SetTileSprite(new VDSprite(graphics.GetDirect3DRenderer(), mapEditorSpriteFileName, 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));
                };

                // make sure we're not still focused on the button
                (void) SetFocus(mapEditor.GetWindow());
            };
        };



        //////////////////////////
        // NPC Editor interface //
        //////////////////////////

        if (npcEditor.GetWindow() != NULL) {

            // Load Sprite button
            /////////////////////

            if (SendMessage(npcEditor.GetLoadSpriteButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char text[FILE_NAME_MAX_SIZE];
                RECT drawingArea;

                // open a sprite image
                VDOpenFile(npcEditor.GetWindow(), "Sprites\0*.PNG\0All Files\0*.*\0", text);

                // make sure the user selected a file
                if (strcmp(text, "") != 0) {
                    // get the sprite's dimensions to pass into VDSprite()
                    (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);

                    // add the sprite file location to the list of them
                    // TODO - recomment
                    //npcEditor.AddNPCSpriteLocation(SendMessage(npcEditor.GetNPCListBox(), LB_GETCOUNT, 0, 0), text);
                    npcEditor.SetNPCSpriteLocation(text);

                    // add the sprite to the editor so that it may be drawn
                    npcEditor.SetNPCSprite(new VDSprite(graphics.GetDirect3DRenderer(), text, 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));
                };

                // make sure we're not still focused on the button
                (void) SetFocus(npcEditor.GetWindow());
            };

            // Save NPC button
            //////////////////

            if (SendMessage(npcEditor.GetSaveNPCButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char npcText[FILE_NAME_MAX_SIZE];

                // check for a name, ID, and sprite
                (void) GetWindowText(npcEditor.GetNameEditBox(), npcText, sizeof(npcText));
                if (strcmp(npcText, "") != 0 && npcEditor.GetNPCSprite() != NULL) {
                    // see if we have an NPC selected
                    if (world.AddNPC(npcText, npcEditor.GetNPCSpriteLocation()) == TRUE) {
                        // update the list box
                        (void) SendMessage(npcEditor.GetNPCListBox(), LB_ADDSTRING, 0, (LPARAM) npcText);
                        //(void) SendMessage(npcEditor.GetNPCListBox(), LB_SETCURSEL, (WPARAM) SendMessage(npcEditor.GetNPCListBox(), LB_GETCOUNT, 0, 0) - 1, 0);

                        // reset the window
                        (void) SendMessage(npcEditor.GetNPCListBox(), LB_SETCURSEL, (WPARAM) - 1, 0);
                        (void) SetWindowText(npcEditor.GetNameEditBox(), "");
                        npcEditor.SetNPCSprite(NULL);
                    };
                };

                // make sure we're not still focused on the button
                (void) SetFocus(npcEditor.GetNPCListBox());
            };

            // Remove NPC button
            ////////////////////

            if (SendMessage(npcEditor.GetRemoveNPCButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                char npcText[FILE_NAME_MAX_SIZE];

                // get the index of the selected NPC text
                LRESULT index = SendMessage(npcEditor.GetNPCListBox(), LB_GETCURSEL, 0, 0);
                if (index != LB_ERR && index >= 0) {
                    // get the name of the selected NPC
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_GETTEXT, (WPARAM) index, (LPARAM) &npcText);

                    // remove the NPC from the world
                    world.RemoveNPC(npcText);

                    // update the list box
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_DELETESTRING, (WPARAM) index, 0);

                    // reset the window
                    (void) SendMessage(npcEditor.GetNPCListBox(), LB_SETCURSEL, (WPARAM) - 1, 0);
                    (void) SetWindowText(npcEditor.GetNameEditBox(), "");
                    npcEditor.SetNPCSprite(NULL);
                    (void) SendMessage(npcEditor.GetDialogueListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");
                };

                // make sure we're not still focused on the button
                (void) SetFocus(npcEditor.GetWindow());
            };

            // user has an NPC selected
            ///////////////////////////

            int currentNPCIndex = SendMessage(npcEditor.GetNPCListBox(), LB_GETCURSEL, 0, 0);
            if (currentNPCIndex != LB_ERR) {
                char currentNPCSelection[FILE_NAME_MAX_SIZE];

                // get the name of the current selected NPC
                (void) strcpy_s(currentNPCSelection, "");
                (void) SendMessage(npcEditor.GetNPCListBox(), LB_GETTEXT, (WPARAM) currentNPCIndex, (LPARAM) &currentNPCSelection);

                // user selects a different NPC
                ///////////////////////////////

                if (strcmp(npcEditorPreviousNPCSelection, currentNPCSelection) != 0) {
                    RECT drawingArea;

                    // copy the current selection to previous
                    (void) strcpy_s(npcEditorPreviousNPCSelection, currentNPCSelection);

                    // update the window
                    (void) SetWindowText(npcEditor.GetNameEditBox(), currentNPCSelection);
                    (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);
                    npcEditor.SetNPCSprite(new VDSprite(graphics.GetDirect3DRenderer(), world.GetNPC(currentNPCSelection).GetImageFileLocation(), 32, 32, D3DXVECTOR3(0.0f, 0.0f, 0.0f), D3DXVECTOR3((float) drawingArea.left, (float) drawingArea.top, 0.0f), D3DCOLOR_ARGB(255, 255, 255, 255), FALSE, FALSE));

                    // reset the window
                    (void) SendMessage(npcEditor.GetDialogueListBox(), LB_RESETCONTENT, 0, 0);
                    (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                    (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");

                    // populate the dialogue list
                    if (SendMessage(npcEditor.GetDialogueListBox(), LB_GETCOUNT, 0, 0) == 0) {
                        // see how many dialogues the NPC has
                        int numberOfDialogues = world.GetNPC(currentNPCSelection).GetDialoguesSize();

                        // add the dialogues to the list
                        while (numberOfDialogues > 0) {
                            char dialogue[FILE_NAME_MAX_SIZE];

                            // set it up so it's "ID: Text"
                            (void) _itoa_s(world.GetNPCDialogueTextID(currentNPCSelection, numberOfDialogues - 1), dialogue, FILE_NAME_MAX_SIZE, 10);
                            (void) strcat_s(dialogue, ": ");
                            (void) strcat_s(dialogue, world.GetNPCDialogueText(currentNPCSelection, numberOfDialogues - 1));
                            (void) SendMessage(npcEditor.GetDialogueListBox(), LB_ADDSTRING, 0, (LPARAM) dialogue);
                            numberOfDialogues--;
                        };
                    };
                };

                // Save Text button
                ///////////////////

                if (SendMessage(npcEditor.GetSaveTextButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                    char text[FILE_NAME_MAX_SIZE], textID[FILE_NAME_MAX_SIZE];

                    // check for text and ID
                    (void) GetWindowText(npcEditor.GetTextEditBox(), text, sizeof(text));
                    (void) GetWindowText(npcEditor.GetTextIDEditBox(), textID, sizeof(textID));
                    if (strcmp(text, "") != 0 && strcmp(textID, "") != 0) {
                        // add the dialogue
                        (void) world.AddNPCDialogue(currentNPCSelection, atoi(textID), text);

                        // reset the window
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_RESETCONTENT, 0, 0);
                        (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                        (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");

                        // populate the dialogue list
                        if (SendMessage(npcEditor.GetDialogueListBox(), LB_GETCOUNT, 0, 0) == 0) {
                            // see how many dialogues the NPC has
                            int numberOfDialogues = world.GetNPC(currentNPCSelection).GetDialoguesSize();

                            // add the dialogues to the list
                            while (numberOfDialogues > 0) {
                                char dialogue[FILE_NAME_MAX_SIZE];

                                // set it up so it's "ID: Text"
                                (void) _itoa_s(world.GetNPCDialogueTextID(currentNPCSelection, numberOfDialogues - 1), dialogue, FILE_NAME_MAX_SIZE, 10);
                                (void) strcat_s(dialogue, ": ");
                                (void) strcat_s(dialogue, world.GetNPCDialogueText(currentNPCSelection, numberOfDialogues - 1));
                                (void) SendMessage(npcEditor.GetDialogueListBox(), LB_ADDSTRING, 0, (LPARAM) dialogue);
                                numberOfDialogues--;
                            };
                        };
                    };

                    // make sure we're not still focused on the button
                    SetFocus(npcEditor.GetWindow());
                };

                // Remove Text button
                /////////////////////

                if (SendMessage(npcEditor.GetRemoveTextButton(), BM_GETSTATE, 0, 0) == BST_FOCUS) {
                    char text[FILE_NAME_MAX_SIZE];

                    // get the index of the selected Dialogue text
                    LRESULT index = SendMessage(npcEditor.GetDialogueListBox(), LB_GETCURSEL, 0, 0);
                    if (index != LB_ERR && index >= 0) {
                        char id[10];

                        // get the name of the selected dialogue
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_GETTEXT, (WPARAM) index, (LPARAM) &text);

                        // remove the dialogue from the NPC
                        int n = 0;
                        while (text[n] != ':') {
                            id[n] = text[n];
                            n++;
                        };
                        id[n] = '\0';
                        world.RemoveNPCDialogue(currentNPCSelection, atoi(id));

                        // update the list box
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_DELETESTRING, (WPARAM) index, 0);

                        // reset the window
                        (void) SendMessage(npcEditor.GetDialogueListBox(), LB_SETCURSEL, (WPARAM) - 1, 0);
                        (void) SetWindowText(npcEditor.GetTextEditBox(), "");
                        (void) SetWindowText(npcEditor.GetTextIDEditBox(), "");
                    };

                    // make sure we're not still focused on the button
                    (void) SetFocus(npcEditor.GetWindow());
                };

                // user has a dialogue selected
                ///////////////////////////////

                int currentDialogueIndex = SendMessage(npcEditor.GetDialogueListBox(), LB_GETCURSEL, 0, 0);
                if (currentDialogueIndex != LB_ERR) {
                    char currentDialogueSelection[FILE_NAME_MAX_SIZE];

                    // get the name of the current selected dialogue
                    (void) strcpy_s(currentDialogueSelection, "");
                    (void) SendMessage(npcEditor.GetDialogueListBox(), LB_GETTEXT, (WPARAM) currentDialogueIndex, (LPARAM) &currentDialogueSelection);

                    // user selects a different NPC
                    if (strcmp(npcEditorPreviousDialogueSelection, currentDialogueSelection) != 0) {
                        char id[10];

                        // copy the current selection to previous
                        (void) strcpy_s(npcEditorPreviousDialogueSelection, currentDialogueSelection);

                        // remove the dialogue from the NPC
                        int n = 0;
                        while (currentDialogueSelection[n] != ':') {
                            id[n] = currentDialogueSelection[n];
                            n++;
                        };
                        id[n] = '\0';

                        // update the window
                        (void) SetWindowText(npcEditor.GetTextEditBox(), world.GetNPCDialogueTextFromID(currentNPCSelection, atoi(id)));
                        (void) SetWindowText(npcEditor.GetTextIDEditBox(), id);

                        // populate the dialogue list
                        if (SendMessage(npcEditor.GetDialogueListBox(), LB_GETCOUNT, 0, 0) == 0) {
                            // see how many dialogues the NPC has
                            int numberOfDialogues = world.GetNPC(currentNPCSelection).GetDialoguesSize();

                            // add the dialogues to the list
                            while (numberOfDialogues > 0) {
                                char dialogue[FILE_NAME_MAX_SIZE];

                                // set it up so it's "ID: Text"
                                (void) _itoa_s(world.GetNPCDialogueTextID(currentNPCSelection, numberOfDialogues - 1), dialogue, FILE_NAME_MAX_SIZE, 10);
                                (void) strcat_s(dialogue, ": ");
                                (void) strcat_s(dialogue, world.GetNPCDialogueText(currentNPCSelection, numberOfDialogues - 1));
                                (void) SendMessage(npcEditor.GetDialogueListBox(), LB_ADDSTRING, 0, (LPARAM) dialogue);
                                numberOfDialogues--;
                            };
                        };
                    };
                };
            };
        };



        // movement (via arrow keys)
        ////////////////////////////

        // see if any keys were pressed
        if (input.GetKeyboardInput()[DIK_ESCAPE] & 0x80) {
            mapMakerRunning = FALSE;
        } else if (input.GetKeyboardInput()[DIK_LEFT] & 0x80) {
            choice = 1;
        } else if (input.GetKeyboardInput()[DIK_RIGHT] & 0x80) {
            choice = 2;
        } else if (input.GetKeyboardInput()[DIK_UP] & 0x80) {
            choice = 3;
        } else if (input.GetKeyboardInput()[DIK_DOWN] & 0x80) {
            choice = 4;
        };

        // user decided to move
        if (choice >= 1 && choice <= 4) {
            // update the user's location
            user.Move(choice);
            choice = 0;
        };



        // world editing
        ////////////////

        // cursor is on the map area
        if (actualMousePositionX >= 0 && actualMousePositionY < width) {
            // user clicks on the map
            if (input.GetMouseInput().rgbButtons[0] & 0x80) {
                // a sprite is loaded
                if (mapEditor.GetTileSprite() != NULL) {
                    int worldX = 0;
                    int worldY = 0;

                    // get the current coordinates relative to the user's position (when they move with thearrow keys)
                    float onScreenX = (float) user.GetCurrentLocationX() * 32.0f + actualMousePositionX - (actualMousePositionX % 32);
                    float onScreenY = (float) user.GetCurrentLocationY() * 32.0f + actualMousePositionY - (actualMousePositionY % 32);

                    // set the world coordinates in relation to the screen coordinates
                    if (world.GetWorldSizeX() == 0 && world.GetWorldSizeY() == 0) {
                        // first tile of the map, set the offset of the world coordinates
                        worldOffset.SetCurrentLocation((int) onScreenX / 32, (int) onScreenY / 32);
                    } else {
                        worldX = (int) onScreenX / 32 - worldOffset.GetCurrentLocationX();
                        worldY = (int) onScreenY / 32 - worldOffset.GetCurrentLocationY();
                    };

                    // TODO: REMOVE
                    graphics.RemoveTexts();
                    char testess[100];
                    sprintf_s(testess, "World Size: %d, %d\n", world.GetWorldSizeX(), world.GetWorldSizeY());
                    graphics.AddText(testess, 10, 10, 80, 18, D3DCOLOR_ARGB(255, 255, 255, 255));
                    char worldCoord[100];
                    sprintf_s(worldCoord, "World Position: %d, %d\n", worldX, worldY);
                    graphics.AddText(worldCoord, 10, 30, 80, 48, D3DCOLOR_ARGB(255, 255, 255, 255));

                    // see if the tile is out of the current world bounds
                    if (world.CheckNewLocation(0, worldX, worldY) == NULL) {
                        // if so, reallocate the world
                        int method = world.ReallocateWorld(worldX, worldY);

                        // we have to make sure the world view is offset
                        switch (method) {
                            case 1:
                                worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX() + worldX,
                                                               worldOffset.GetCurrentLocationY());
                                break;
                            case 2:
                                worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX(),
                                                               worldOffset.GetCurrentLocationY() + worldY);
                                break;
                            case 3:
                                worldOffset.SetCurrentLocation(worldOffset.GetCurrentLocationX() + worldX,
                                                               worldOffset.GetCurrentLocationY() + worldY);
                                break;
                            default:
                                break;
                        };
                    };

                    // the tile was in the negative world space, so it's 0 in the new world space
                    if (worldX < 0) {
                        worldX = 0;
                    };
                    if (worldY < 0) {
                        worldY = 0;
                    };

                    // set the tile status
                    if (SendMessage(mapEditor.GetNullTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 0);
                    } else if (SendMessage(mapEditor.GetOpenTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 1);
                    } else if (SendMessage(mapEditor.GetImpassableTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 2);
                    } else if (SendMessage(mapEditor.GetNewAreaTileOption(), BM_GETCHECK, 0, 0) == BST_CHECKED) {
                        world.SetTileStatus(worldX, worldY, 3);
                    } else {
                        world.SetTileStatus(worldX, worldY, 0);
                    };

                    // set the name of the texture in world
                    world.SetTileImageFileLocation(worldX, worldY, mapEditorSpriteFileName);

                    // check if there's a tile already in that spot and remove it if there is
                    graphics.RemoveTileSprite(onScreenX, onScreenY);

                    // add the new sprite
                    graphics.AddSprite(mapEditorSpriteFileName, onScreenX, onScreenY, 32, 32, FALSE, FALSE);
                };
            };
        };



        // drawing
        //////////

        // draw graphics
        renderDimensions.left = 0;
        renderDimensions.top = 0;
        renderDimensions.right = width - GetSystemMetrics(SM_CXBORDER);
        renderDimensions.bottom = height - GetSystemMetrics(SM_CYBORDER);
        graphics.Render(&renderDimensions, 0, 0, user.GetCurrentLocationX(), user.GetCurrentLocationY());

        // draw Map Editor sprites
        if (mapEditorWindow != NULL) {
            RECT drawingArea;
            (void) mapEditor.GetTileSpriteDimensions(&drawingArea);
            graphics.Render(&drawingArea, mapEditor.GetWindow(), mapEditor.GetTileSprite());
        };

        // draw NPC Editor sprites
        if (npcEditorWindow != NULL) {
            RECT drawingArea;
            (void) npcEditor.GetNPCSpriteDimensions(&drawingArea);
            graphics.Render(&drawingArea, npcEditor.GetWindow(), npcEditor.GetNPCSprite());
        };
    };
};



/* **********************************************************************************************************************
   RegisterWindows
   ---------------
   Registers the window classes that we'll be using in this application.
   #param  instance   Instance of the application.
   **********************************************************************************************************************/

void VDEngine::RegisterWindows(HINSTANCE instance) {
    WNDCLASSEX windowClassStructure;
    ATOM registerClassEx;

    // main window
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_ENGINE_CLASS_NAME;             // class name
    windowClassStructure.lpszMenuName = MAKEINTRESOURCE(VD_MENU);          // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };
    (void) SecureZeroMemory(&windowClassStructure, sizeof(windowClassStructure));

    // Map Editor
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_MAP_EDITOR_CLASS_NAME;         // class name
    windowClassStructure.lpszMenuName = NULL;                              // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };

    // NPC Editor
    windowClassStructure.cbClsExtra = 0;                                   // number of extra bytes to allocate following the window-class structure
    windowClassStructure.cbSize = sizeof(WNDCLASSEX);                      // size of the structure (in bytes)
    windowClassStructure.cbWndExtra = 0;                                   // number of extra bytes to allocate following the window instance
    windowClassStructure.hbrBackground = GetSysColorBrush(COLOR_BTNFACE);  // background brush
    windowClassStructure.hCursor = LoadCursor(NULL, IDC_ARROW);            // cursor
    windowClassStructure.hIcon = NULL;                                     // icon
    windowClassStructure.hIconSm = LoadIcon(NULL, IDI_APPLICATION);        // small icon
    windowClassStructure.hInstance = instance;                             // instance that has the window procedure
    windowClassStructure.lpfnWndProc = WindowProcedure;                    // window procedure
    windowClassStructure.lpszClassName = VD_NPC_EDITOR_CLASS_NAME;         // class name
    windowClassStructure.lpszMenuName = NULL;                              // menu resource name
    windowClassStructure.style = NULL;                                     // style
    registerClassEx = RegisterClassEx(&windowClassStructure);
    if (registerClassEx == 0) {
        VDError("Failed to register the window class.", __FILE__, __LINE__);
    };
};



/* **********************************************************************************************************************
   CreateVDEngineWindow
   --------------------
   Creates the main window of the engine.
   #param  instance   Instance of the application.
   #param  className  The name of the window class.
   #param  width      Width of the window.
   #param  height     Height of the window.
   #return  A handle to the window.
   **********************************************************************************************************************/

HWND VDEngine::CreateVDEngineWindow(HINSTANCE instance, LPCSTR className, int width, int height) {
    // create the menu resource
    HMENU menu = LoadMenu(instance, MAKEINTRESOURCE(VD_MENU));
    if (menu == NULL) {
        VDError("Failed to create the menu resource.", __FILE__, __LINE__);
    };

    return CreateWindowEx(NULL,                                                   // extended window styles
                          className,                                              // class name
                          "VDEngine",                                             // window name
                          WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_VISIBLE,  // window styles
                          0,                                                      // x coordinate (0 = left)
                          0,                                                      // y coordinate (0 = top)
                          width,                                                  // width
                          height,                                                 // height
                          NULL,                                                   // parent/owner window
                          menu,                                                   // menu
                          instance,                                               // instance
                          NULL);                                                  // used for CREATESTRUCT structure
};



//  TODO

void VDEngine::VDOpenFile(HWND window, LPSTR filter, LPSTR fileName) {
    char file[FILE_NAME_MAX_SIZE];
    OPENFILENAME openFileBox;

    // zero out the open file box just to be on the safe side
    SecureZeroMemory(&openFileBox, sizeof(OPENFILENAME));

    // set up the open file structure
                openFileBox.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_NOCHANGEDIR;
            openFileBox.hwndOwner = window;
            openFileBox.lpstrFile = file;
         openFileBox.lpstrFile[0] = '\0';
       openFileBox.lpstrFileTitle = NULL;
          openFileBox.lpstrFilter = filter;
      openFileBox.lpstrInitialDir = NULL;
          openFileBox.lStructSize = sizeof(OPENFILENAME);
         openFileBox.nFilterIndex = 1;
             openFileBox.nMaxFile = sizeof(file);
        openFileBox.nMaxFileTitle = NULL;

    // open a file
    BOOL returnValue = GetOpenFileName(&openFileBox);
    if (returnValue != 0) {
        (void) strcpy_s(fileName, sizeof(file), file);
    } else {
        (void) strcpy_s(fileName, sizeof(file), "");
    };
};